#!/usr/bin/env python3
""" Copyright 2020 Phil Chandler
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""novag_xboard.py v1.5
    This script is designed to act as a pseudo engine in Xboard, so that you can play against a Novag Sapphire, Diamond and later models
    using the Xboard interface.  It has been tested against a Sapphire, Diamond II, Star Diamond, Super Nova and Super Expert C (via the
    MAME emulation). It is NOT presently working for the Citrine.  It can be easily modified
    to work with other models such as the Super VIP, with suitable work arounds for variations in the Novag serial protocol.
    
    Prerequisites are Python3 and the PySerial library. Since this is written in Python, theoretically, it
    should work under Windows, but that hasn't been tested. In Windows the built in support for Novag computers in Arena is recommended,
    and the UCI engines created by Alain Zanchetta: http://alain.zanchetta.free.fr/softs/Alainza.Uci.1.0.335.1.Setup.msi
    
    For a Sapphire or Diamond you will need a suitable hardware adapter to connect to a
    PC serial port, or one of the original Novag Super System Distributor hubs.  You can buy a modern adapter from https://sites.google.com/site/bergersprojects/chess_accessories/novag-arena_interface or build
    one yourself following instructions from https://www.schach-computer.info/wiki/index.php?title=Novag_Distributor_Replacement.
    
    This script was developed on a PC running Linux (Ubuntu 20.04 as it happens). Different generations of machines have 
    slight differences in their serial protocol, so some adjustments are needed. The Sapphire doesn't, for example, support
    the 'J' command listed in the documentation to tell the Novag to 'Go'. You need to use the actual Go button to play black, or swap
    colours mid-game.  Essentially you tell Xboard the engine is now playing the other colour and then press Go on the Novag.

    The script takes the serial port as a command line argument, and then works out what baud rate to use for the computer connected.
    If you want to play against a computer you add it as the 1st engine in Xboard, giving the command for the engine as something like:

    "<path to script>/novag_xboard.py /ttyUSB0"

    The script will get the model name from the computer and tell Xboard to use that as the machine's nickname.

    For computer vs computer games, (assuming you have a second serial port available), a good way of starting Xboard is with a command
    something like this:

    xboard -fcp "<path to script>/novag_xboard.py /dev/ttyUSB0" -fd "/<path to script>" -scp "<path to script>/novag_xboard.py /dev/ttyUSB1" -sd "<path to script>"

    -fcp/-fd define the parameters for the 1st engine and -scp/-sd for the second.  To change the colours, swap the serial ports.

    Start the game by selecting 'Two Machines' from the Xboard Mode menu.
    
    Any bug reports, improvements or feedback, please send to user kalatov at https://hiarcs.net/forums/ For outputing detailed debugging
    information see the notes below"""

import serial, time, sys, string
# Set debug = True to send additional output to Xboard for logging in debug mode
# debug2 = True gives more indepth information
# To see debug info run Xboard with the command 'xboard -debug -engineDebugOutput 2'
# and tail -f the xboard.debug file in the directory you are running the script from.
# There is also logging to a file which has the advantage of capturing information which
# may not make it as fas as xboard.
logfile = "/home/chandler/novag_xboard.log"
log = open(logfile, "a")
debug = True
debug2 = True
name = ""
status = "new"
colour = "black" #default unless changed later
ucbmode = False # whether computer is acting as an electronic board
port = str(sys.argv[1])
if debug2:
    print(port)
baudrate = 57600 # set as default, the script will check if this works later

def send_instruction(instruction, **kwargs):
    if debug:
        print("Instruction to be sent to computer: " + instruction)
        log.write("Instruction to be sent to computer: " + instruction)
        log.write("\n")
        log.flush()
    if debug2:
        print(ser)
        log.write(str(ser))
        log.write("\n")
        log.flush()
    compresponse = ""
    if ucbmode:
        ser.write(("U on" + '\r\n').encode('utf-8')) # if acting as a board, enforce referee mode
        print(ser.read_all())
    ser.write((instruction + '\r\n').encode('utf-8')) # Send the instruction
    while ser.in_waiting == 0:
        waiting = True
    moredata = True
    while moredata:
        response = ser.read_until() # Get the computer's response
        responsestr =str(response)
        if debug:
            print("raw response from computer: " + responsestr)
            log.write("raw response from computer: " + responsestr)
            log.write("\n")
            log.flush()
        checkendcode = check_endcode(responsestr) #Is there a resignation, mate etc.
        if checkendcode == 'nocode':
            if instruction[0] == 'M' or instruction[0] == 'k' or instruction[0] == 'J': #It's a move or command to start game
                if "UCBII" in name: #Computer is a Citrine
                    ser.write(('\r\n').encode('utf-8')) # Send an extra line feed otherwise Citrine waits for piece to be moved.
                if instruction[0] == 'M': #it's a move, consume echo and get computer's move
                    response = ser.read_until()
                responsearray = response.split()
                responsestr =str(responsearray)
                if debug:
                    print("raw response from computer: " + responsestr)
                    log.write("raw response from computer: " + responsestr)
                    log.write("\n")
                    log.flush()
                if debug2:
                    print("Responsearray:")
                    print(responsearray)
                    log.write("Responsearray:")
                    log.write(str(responsearray))
                    log.write("\n")
                    log.flush()
                checkendcode = check_endcode(responsestr) #Is there a resignation, mate etc.
                if checkendcode == 'nocode':
                    for element in responsearray:
                        move = ""
                        #checkendcode = check_endcode(str(element))
                        movechars = ["-","x","+","q","n"]
                        for char in movechars:
                            if char in str(element): #Look for the element that actually has the move
                                move1 = str(element)
                                move2 = move1.split('\'')
                                if move2[1][0] == ',': #Check if we're talking to e.g. a Super Expert C
                                    move3 = move2[1].lstrip(',')
                                    move = move3.rstrip('|')
                                else:
                                    move = move2[1]
                                compresponse = "move " + move
            elif instruction[0] == 'L': #We're setting the level so need to consume the extra line returned
                response = ser.read_until()
                compresponse =str(response)
            elif instruction[0] == 'I':
                if "ID." in responsestr:
                    idarray = responsestr.split('ID. ') #Star Diamond variation
                else:
                    idarray = responsestr.split('ID ')
                idarray2 = idarray[1].split("\\r")
                compresponse = idarray2[0].strip()
            if debug:
                print("Response from computer: " + compresponse)
                log.write("Response from computer: " + compresponse)
                log.write("\n")
                log.flush()
            print(compresponse)
        if debug2:
            log.write("Data in buffer: " + str(ser.in_waiting))
            log.write("\n")
            log.flush()
        if ser.in_waiting == 0:
            moredata = False
    return compresponse

def set_option(xbinput):
    optionarray = xbinputarray[1].split("=")
    option = optionarray[0]
    value = optionarray[1]
    if option == "Level":
        response = send_instruction("L "+ value)
        optionset = "Level set to " + response[10 : 14]
    # Add other options as elseifs, ending by setting optionset
    # Options are defined in the 'feature' line sent in response to 'protover' from Xboard.
    return optionset

def send_message(message):
    print("telluser " + message)
    return

def check_endcode(responsestr):
    if debug2:
        print("Checking for endcode in: " + responsestr)
        log.write("Checking for endcode in: " + responsestr)
        log.write("\n")
        log.flush()
    checkendcode = responsestr.find("M#")
    if debug2:
        print("Checkendcode: " + str(checkendcode))
        log.write("Checkendcode: " + str(checkendcode))
        log.write("\n")
        log.flush()
    if checkendcode != -1: # It's a code for how the game ended
        message = responsestr[checkendcode : checkendcode + 3]
        if debug:
            print("message code: " + message)
            log.write("message code: " + message)
            log.write("\n")
            log.flush()
        if message == 'M#1':
            result = "1/2-1/2 {Draw by repetition}"
        elif message == 'M#2':
            result = "1/2-1/2 {Draw by 50 move rule}"
        elif message == 'M#3':
            result = "1/2-1/2 {Draw by insufficient material}"
        elif message == 'M#4':
            result = "1/2-1/2 {Draw by stalemate}"
        elif message == 'M#5':
            if colour == "white":
                result = "1-0 {white mates}"
            elif colour == "black":
                result = "0-1 {black mates}"
        elif message == 'M#6':
            if colour == "white":
                result = "0-1 {white resigns}"
            elif colour == "black":
                result = "1-0 {black resigns}"
    else:
        result = "nocode"
    if result != "nocode":
        if debug:
            log.write("Result: " + result)
            log.write("\n")
            log.flush()
        print(result)
    return result


ser = serial.Serial(port, baudrate, timeout=3.0)  # open serial port
#Test if we have communications
if debug2:
    if ser.is_open:
        print("Port is open")
        print(ser)
# Switch Xmit on and get the device name
ser.write(("X on" + '\r\n').encode('utf-8'))
response = ser.read_until() # Get the computer's response
commstest =str(response)
if commstest != "b'Xmit on\\r\\n'": #57600 didn't work, try 9600 (Trying 9600 first breaks Star Diamonds)
    ser.close()
    ser.baudrate = 9600
    ser.open()
    if debug2:
        print(ser)
        if ser.is_open:
            print("Port is open")
    # This is probably a Citrine that takes a lot to wake it up
    ser.write(("X on" + '\r\n').encode('utf-8'))
    ser.read_until()
    ser.write(("X on" + '\r\n').encode('utf-8'))
    ser.read_until()
    ser.write(("X on" + '\r\n').encode('utf-8'))
    response = ser.read_until() # Get the computer's response
    commstest =str(response)
    if commstest != "b'Xmit on\\r\\n'":
        print("No computer detected")
        ser.reset_input_buffer()
        ser.reset_output_buffer()
        ser.close()
        exit()
name = send_instruction("I")
ser.timeout = None #Timeout can cause missed moves 

# Wait for input from Xboard
xbinput = ""
while 1 :
    xbinput = input()
    xbinputarray = xbinput.split()
    command = xbinputarray[0]
    if debug == True:
        print("Command from Xboard was " + xbinput)
        log.write("Command from Xboard was " + xbinput)
        log.write("\n")
        log.flush()
    if command == 'quit':
        #send_instruction("new")
        ser.reset_input_buffer()
        ser.reset_output_buffer()
        ser.close()
        exit()
    elif command == 'xboard':
        print("\n")
    elif command == 'protover':
        # Set which features accepted and define any engine options configurable in the engine options dialogue of the GUI.
        # The option defined here is the ability to set the computers level, useing the EA1, TR1 etc. notation.
        print("feature usermove=1 ping=1 time=0 draw=0 sigint=0 sigterm=0 option=\"Level -string e.g. TR1\" myname=\""  + name +"\" colors=1 reuse=0")
        print("feature done=1")
    elif command == 'new':
        send_instruction(command)
        print("new")
        status = 'new'
    elif command == 'black':
        colour = "black"
    elif command == 'white':
        colour = "white"
    elif command == 'go':
        if status == 'new':
            if "DIAMOND II" in name or "H8S/SD" in name or "UCBII" in name:
                send_instruction("J")
            elif name == "E3.6":
                send_instruction("kc7")
            else:
                send_instruction("k90")
            status = 'started'
    elif command == 'usermove':
        movestr = xbinputarray[1]
        if debug:
            print("Received move from xbd: " + movestr)
            log.write("Received move from xbd: " + movestr)
            log.write("\n")
            log.flush()
        if len(movestr) > 4:
            if debug2:
                print("Length of move string: " + str(len(movestr)))
                log.write("Length of move string: " + str(len(movestr)))
                log.write("\n")
                log.flush()
            if movestr[4].isalpha: #Do we have a promotion?
                movestrtmp = movestr[:4] + "/" + movestr[4]
                movestr = movestrtmp
        if debug:
            print("Movestr to be sent to computer: " + movestr)
            log.write("Movestr to be sent to computer: " + movestr)
            log.write("\n")
            log.flush()
        instruction = "M" + movestr
        send_instruction(instruction, colour=colour)
        if status == 'new':
            status = 'started'
    elif command == 'remove':
        send_instruction("T")
        send_instruction("T")
    elif command == 'option':
        set_option(xbinput)
    elif command == 'I':
        send_instruction("I")
    elif command == 'accepted' or command == 'rejected':
        True #No response
    elif command == 'result':
        # game over, let's switch to Ref mode and flush the serial buffers
        send_instruction("U on")
        ser.reset_input_buffer()
        ser.reset_output_buffer()
    elif command == 'ucbmode':
        ucbmode = True
    elif command == 'ping':
        print("pong " + xbinputarray[1])
    else:
        print("Error (unknown command): " + xbinput)